from typing import Any
from pydantic import Field

from muck_out.types.common import CommonObject

from .common import Common


class Object(Common, CommonObject):
    attributed_to: str = Field(
        examples=["https://actor.example/"],
        description="id of the actor that authored this object",
        alias="attributedTo",
    )
    content: str | None = Field(default=None, description="The content of the object")

    attachment: list[dict[str, Any]] | None = Field(
        None,
        description="A list of objects that are attached to the original object",
    )
    tag: list[dict[str, Any]] | None = Field(
        None,
        description="A list of objects that expand on the content of the object",
    )
    url: list[dict[str, Any]] = Field(
        default=[],
        description="A list of urls that expand on the content of the object",
    )
    sensitive: bool | None = Field(
        None,
        description="""
    Marks the object as sensitive. Currently, used by everyone, a better way would be an element of the tag list that labels the object as sensitive due a reason
    """,
    )



from pydantic import BeforeValidator
from typing import Annotated, Any

from bovine.activitystreams.utils import id_for_object
from muck_out.transform.utils import sanitize_html
from muck_out.transform import transform_to_list_of_uris, transform_url


def safe_html(x) -> str | None:
    if x is None:
        return None
    if isinstance(x, list):
        if len(x) == 0:
            return None
        x = x[0]
    return sanitize_html(x)


def safe_id_for_object(x):
    if isinstance(x, list):
        if len(x) == 0:
            return None
        x = x[0]
    return id_for_object(x)


HtmlStringOrNone = Annotated[str | None, BeforeValidator(safe_html)]
"""Used for strings, which may contain html"""

IdFieldOrNone = Annotated[str | None, BeforeValidator(safe_id_for_object)]
"""Used for fields meant to contain an id"""


UrlList = Annotated[list[dict[str, Any]], BeforeValidator(transform_url)]
"""Transforms a list of urls"""

TransformToListOfUris = Annotated[list[str], BeforeValidator(transform_to_list_of_uris)]
"""Transforms a list of recipients, ensuring it is a list of URIs"""



from bovine.activitystreams.utils import id_for_object, uris_for_public


from .list_utils import transform_to_list

default_public = "https://www.w3.org/ns/activitystreams#Public"
"""Default URI to represent something is being addressed to the public."""


def transform_url_part(url_part):
    if isinstance(url_part, str):
        url_part = {"type": "Link", "href": url_part}
    return url_part


def transform_url(url) -> list[dict] | None:
    """Transform to a list of links

    ```python
    >>> transform_url("http://remote.test/html")
    [{'type': 'Link', 'href': 'http://remote.test/html'}]

    ```
    """
    url_list = transform_to_list(url)

    return [transform_url_part(url_part) for url_part in url_list]


def normalize_public(uri: str) -> str:
    """
    Normalizes the representation of public, i.e.

    ```python
    >>> normalize_public("as:Public")
    'https://www.w3.org/ns/activitystreams#Public'

    ```
    """
    if uri in uris_for_public:
        return default_public
    return uri


def transform_to_list_of_uris(data) -> list[str]:
    """Transforms to a list of URIs. As this simplifies a
    complicated data structure, some information may be lost.

    ``` python
    >>> transform_to_list_of_uris("http://remote.test/id")
    ['http://remote.test/id']

    >>> transform_to_list_of_uris({"id": "http://remote.test/id"})
    ['http://remote.test/id']

    >>> transform_to_list_of_uris({"ignored": True})
    []

    ```
    """
    to_list = transform_to_list(data)

    return [normalize_public(x) for x in (id_for_object(to) for to in to_list) if x]



from pydantic import Field
from .common import CommonAll


class Collection(CommonAll):
    """Abstracts all the ActivityPub collection concepts"""

    type: str = Field(
        examples=[
            "Collection",
            "OrderedCollection",
            "CollectionPage",
            "OrdererCollectionPage",
        ],
        description="""Type of object""",
    )

    items: list[str] | None = Field(None, description="""The items""")

    next: str | None = Field(None)
    prev: str | None = Field(None)
    first: str | None = Field(None)
    last: str | None = Field(None)
    total_items: int | None = Field(None, alias="totalItems")



import pytest
from faststream.rabbit import RabbitBroker
from faststream.rabbit.testing import TestRabbitBroker

from cattle_grid.dependencies.globals import global_container

from muck_out.testing.examples import actor

from . import FetchActor


@pytest.fixture
def transform_then_call():
    async def func(method):
        broker = RabbitBroker()

        broker.subscriber("queue")(method)

        @broker.subscriber("fetch", exchange=global_container.exchange)
        async def fetch():
            return {"parsed": {"actor": actor}}

        async with TestRabbitBroker(broker) as br:
            await br.publish({}, "queue")

    return func


async def test_fetch_actor(transform_then_call):
    async def method(fetch: FetchActor):
        result = await fetch("http://actor.test", "http://object.test")
        assert result

    await transform_then_call(method)



from . import normalize_data as normalize


def test_normalize():
    normalize({})


def test_normalize_activity():
    result = normalize(
        {
            "id": "http://actor.example/2",
            "type": "Create",
            "to": ["http://remote.example"],
            "actor": "http://actor.example/",
            "object": {
                "type": "Note",
                "id": "http://actor.example/1",
                "to": ["http://remote.example"],
                "attributedTo": "http://actor.example/",
                "content": "moo",
            },
        }
    )

    assert result.activity
    assert result.embedded_object



from typing import Any
import pytest

pytest.importorskip("cattle_grid")

from faststream.rabbit import RabbitBroker
from faststream.rabbit.testing import TestRabbitBroker

from cattle_grid.extensions.load import build_transformer
from cattle_grid.model import ActivityMessage
from cattle_grid.testing.fixtures import *  # type: ignore # noqa

from muck_out.extension import extension
from muck_out.types import Activity, Object

from . import ParsedActivity, ParsedActor, ParsedEmbeddedObject


@pytest.fixture
def transform_then_call(actor_for_test):
    async def func(data: dict[str, Any], method):
        transformer = build_transformer([extension])
        transformed = await transformer({"raw": data}, actor_id=actor_for_test.actor_id)  # type: ignore

        broker = RabbitBroker()
        broker.subscriber("queue")(method)

        print(transformed.get("parsed"))

        message = ActivityMessage(actor=actor_for_test.actor_id, data=transformed)
        async with TestRabbitBroker(broker) as br:
            await br.publish(message.model_dump(mode="json"), "queue")

    return func


async def test_injection_activity(transform_then_call):
    data = {
        "@context": "https://www.w3.org/ns/activitystreams",
        "type": "AnimalSound",
        "actor": "http://abel/actor/AFKb0cQunSBv1fC7sWbQYg",
        "to": ["https://www.w3.org/ns/activitystreams#Public"],
        "cc": ["http://abel/followers/RKsezXFc1SGvQKvucioJxg"],
        "published": "2025-09-17T18:34:00Z",
        "content": "meow",
        "id": "http://abel/simple_storage/019958f4-75e2-7039-b3fe-3538d3230d4f",
    }

    def method(activity: ParsedActivity):
        assert isinstance(activity, Activity)
        assert activity.type == "AnimalSound"
        assert activity.content == "meow"

    await transform_then_call(data, method)


async def test_injection_actor_none(transform_then_call):
    data = {
        "@context": "https://www.w3.org/ns/activitystreams",
        "type": "AnimalSound",
        "actor": "http://abel/actor/AFKb0cQunSBv1fC7sWbQYg",
        "to": ["https://www.w3.org/ns/activitystreams#Public"],
        "cc": ["http://abel/followers/RKsezXFc1SGvQKvucioJxg"],
        "published": "2025-09-17T18:34:00Z",
        "content": "meow",
        "id": "http://abel/simple_storage/019958f4-75e2-7039-b3fe-3538d3230d4f",
    }

    def method(actor: ParsedActor):
        assert actor is None

    await transform_then_call(data, method)


async def test_injection_embedded_object(transform_then_call):
    activity = {
        "@context": [
            "https://www.w3.org/ns/activitystreams",
            {"Hashtag": "as:Hashtag", "sensitive": "as:sensitive"},
        ],
        "actor": "http://actor.example",
        "id": "http://actor.example/activity/JTnMd8Pfuuk",
        "object": {
            "attributedTo": "http://actor.example",
            "content": "text",
            "id": "http://actor.example/object/EsHd36ue8zo",
            "published": "2025-09-09T09:32:38Z",
            "to": [
                "https://www.w3.org/ns/activitystreams#Public",
                "http://remote.example/",
            ],
            "type": "Note",
            "url": [
                "http://remote.example/objects/123",
                "http://other.example/objects/123",
            ],
        },
        "published": "2025-09-09T09:32:38Z",
        "to": [
            "https://www.w3.org/ns/activitystreams#Public",
            "http://remote.example/",
        ],
        "type": "Create",
    }

    def method(activity: ParsedActivity, embedded: ParsedEmbeddedObject):
        assert isinstance(activity, Activity)
        assert isinstance(embedded, Object)

    await transform_then_call(activity, method)



import pytest

pytest.importorskip("cattle_grid")

from unittest.mock import MagicMock

from muck_out.testing.examples import actor_example

from .extension import muck_out


async def test_muck_out():
    result = await muck_out(
        {"raw": actor_example}, actor=MagicMock(actor_id="http://actor.test/")
    )

    assert "parsed" in result


async def test_muck_out_mastodon_like():
    like = {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": "https://mastodon.social/users/the_milkman#likes/251507741",
        "type": "Like",
        "actor": "https://mastodon.social/users/the_milkman",
        "object": "https://dev.bovine.social/html_display/object/01999580-e682-799a-8a43-ae9f5742d148",
    }

    result = await muck_out(
        {"raw": like}, actor=MagicMock(actor_id="http://actor.test/")
    )

    parsed = result.get("parsed", {})

    assert parsed.get("activity")



from cattle_grid.extensions import Extension
from cattle_grid.dependencies.transformer import ActorTransforming

from . import normalize_data

extension = Extension(name="muck out", module=__name__)


@extension.transform(inputs=["raw"], outputs=["parsed"])
async def muck_out(data: dict, actor: ActorTransforming):
    actor_id = actor.actor_id if actor else None
    return {
        "parsed": normalize_data(data["raw"], actor_id=actor_id).model_dump(
            by_alias=True
        )
    }



from uuid6 import uuid7
from bovine.activitystreams.utils import id_for_object

from muck_out.transform.list_utils import list_from_value


def normalize_to(
    value: None | str | list[str], receiving_actor: str | None
) -> list[str]:
    """Normalizes the to value. A normalized activity
    or object should always be addressed to someone.
    If the activity is not addressed to anyone, the receiving
    actor is assumed to be the recipient

    ```pycon
    >>> normalize_to(None, "http://actor.example")
    ['http://actor.example']

    >>> normalize_to("http://to.example", None)
    ['http://to.example']

    >>> normalize_to(["http://alice.example", "http://bob.example"], None)
    ['http://alice.example', 'http://bob.example']

    ```
    """
    if value is None:
        if receiving_actor is None:
            raise ValueError(
                "Cannot set receiving actor as fake recipient if it isn't specified"
            )
        return [receiving_actor]
    return list_from_value(value)  # type:ignore


def normalize_id(activity: dict) -> str:
    """
    Creates a normalized id

    ```pycon
    >>> normalize_id({"id": "http://id.example"})
    'http://id.example'

    >>> normalize_id({})
    Traceback (most recent call last):
        ...
    ValueError: Cannot fake id if actor is not present

    ```
    """
    result = activity.get("id")
    if result is not None:
        return result
    actor_id = id_for_object(activity.get("actor"))

    if actor_id is None:
        raise ValueError("Cannot fake id if actor is not present")

    return f"{actor_id}#fake_id" + str(uuid7())


def normalize_url(url):
    if url is None:
        return
    if isinstance(url, str):
        url = {"type": "Link", "href": url}

    return list_from_value(url)


def normalize_to_id(attributed_to) -> str | None:
    if isinstance(attributed_to, list) and len(attributed_to) == 1:
        attributed_to = attributed_to[0]

    return id_for_object(attributed_to)  # type: ignore



import pytest

from . import ObjectStub


@pytest.mark.parametrize(
    "url,expected",
    [
        ("http://remote.test", [{"type": "Link", "href": "http://remote.test"}]),
        (["http://remote.test"], [{"type": "Link", "href": "http://remote.test"}]),
        (
            [{"type": "Link", "href": "http://remote.test"}],
            [{"type": "Link", "href": "http://remote.test"}],
        ),
        (
            {"type": "Link", "href": "http://remote.test"},
            [{"type": "Link", "href": "http://remote.test"}],
        ),
        (
            ["http://one.test", {"type": "Link", "href": "http://two.test"}],
            [
                {"type": "Link", "href": "http://one.test"},
                {"type": "Link", "href": "http://two.test"},
            ],
        ),
    ],
)
def test_object_stub_url(url, expected):
    stub = ObjectStub(url=url)  # type:ignore

    assert stub.url == expected


def test_object_stub_content():
    stub = ObjectStub(content="content")  # type:ignore

    assert stub.content == "content"


def test_object_parse_emoji():
    data = {
        "type": "Note",
        "attributedTo": "http://actor.example",
        "to": ["https://www.w3.org/ns/activitystreams#Public"],
        "id": "http://actor.example/object/gGhtmEsRxq0",
        "published": "2025-09-09T09:32:39Z",
        "content": "emoji just id :cow:",
        "tag": ["http://pasture-one-actor/assets/cow_emoji.jsonap"],
        "@context": [
            "https://www.w3.org/ns/activitystreams",
            {"Hashtag": "as:Hashtag", "sensitive": "as:sensitive"},
        ],
    }

    stub = ObjectStub.model_validate(data)

    assert stub.tag == [{"href": "http://pasture-one-actor/assets/cow_emoji.jsonap"}]


def test_object_stub_parse_null_url():
    data = {"url": None}
    ObjectStub.model_validate(data)



from muck_out.testing.examples import actor_example, mitra_actor
from muck_out.process import normalize_actor

from . import actor_to_header_info, ActorHeaderInfo


def test_actor_to_header_info():
    actor = normalize_actor(actor_example)
    assert actor

    info = actor_to_header_info(actor)

    assert info.id == actor_example["id"]
    assert info.identifier == "acct:kitty@abel"
    assert info.name == actor_example["name"]
    assert info.avatar_url == "https://dev.bovine.social/assets/bull-horns.png"


def test_actor_to_header_info_html_url():
    actor = normalize_actor(mitra_actor)
    assert actor

    info = actor_to_header_info(actor)
    assert info.html_url == "http://mitra/users/admin"


def test_actor_to_header_info_serialize_and_deserialize():
    actor = normalize_actor(actor_example)
    assert actor

    info = actor_to_header_info(actor)

    dumped = info.model_dump()

    ActorHeaderInfo.model_validate(dumped)



import logging
from typing import Any

from cattle_grid.model.common import WithTransformedData
from muck_out.types import Activity, Actor, Object, Collection

logger = logging.getLogger(__name__)


def data_to_type(data: dict[str, Any], key: str, object_type) -> None:
    try:
        candidate = data.get("parsed", {}).get(key)
        if not candidate:
            return None
        return object_type.model_validate(candidate)
    except Exception as e:
        logger.info(e)
        return None


def get_activity(message: WithTransformedData) -> None | Activity:
    return data_to_type(message.data, "activity", Activity)


def get_actor(message: WithTransformedData) -> None | Actor:
    return data_to_type(message.data, "actor", Actor)


def get_object(message: WithTransformedData) -> None | Object:
    return data_to_type(message.data, "object", Object)


def get_embedded_object(message: WithTransformedData) -> None | Object:
    return data_to_type(message.data, "embedded_object", Object)


def get_collection(message: WithTransformedData) -> None | Collection:
    return data_to_type(message.data, "collection", Collection)



import nh3

from copy import deepcopy


allowed_html_tags = {
    "a",
    "abbr",
    "acronym",
    "b",
    "blockquote",
    "code",
    "em",
    "i",
    "li",
    "ol",
    "strong",
    "ul",
    "p",
    "br",
    "span",
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6",
    "hr",
}
"""The currently allowed list of html tags"""

attributes = deepcopy(nh3.ALLOWED_ATTRIBUTES)

attributes["a"].add("class")


def attribute_filter(tag, attr, value):
    if tag == "a" and attr == "class":
        classes = value.split(" ")
        new_classes = [x for x in classes if x in ["mention", "hashtag"]]
        if len(new_classes) == 0:
            return None
        return " ".join(new_classes)

    return value


def sanitize_html(value: str | None) -> str | None:
    """Cleans html

    ```pycon
    >>> sanitize_html("<p>text</p>")
    '<p>text</p>'

    >>> sanitize_html("<script>alert('xss')</script>")
    ''

    ```
    """
    if isinstance(value, str):
        return nh3.clean(
            value,
            tags=allowed_html_tags,
            attributes=attributes,
            attribute_filter=attribute_filter,
        )
    return None


def remove_html(value: str | None) -> str | None:
    """Removes html

    ```pycon
    >>> remove_html('<a href="http://location.test">location.test</p>')
    'location.test'

    ```
    """

    if isinstance(value, str):
        return nh3.clean(value, tags=set())
    return None



from .activity import activity_stub


def test_mastodon_like():
    mastodon_like = {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": "https://mastodon.social/users/the_milkman#likes/251507741",
        "type": "Like",
        "actor": "https://mastodon.social/users/the_milkman",
        "object": "https://dev.bovine.social/html_display/object/01999580-e682-799a-8a43-ae9f5742d148",
    }

    actor_id = "http://local.test/actor/id"

    result = activity_stub(mastodon_like, actor_id=actor_id)

    assert result.to == [actor_id]



from pydantic import Field
from muck_out.validators import IdFieldOrNone, TransformToListOfUris
from .common import CommonAll


class CollectionStub(CommonAll):
    """Abstracts all the ActivityPub collection concepts"""

    items: TransformToListOfUris = Field([], description="""The items""")

    next: IdFieldOrNone = Field(None)
    prev: IdFieldOrNone = Field(None)
    first: IdFieldOrNone = Field(None)
    last: IdFieldOrNone = Field(None)
    total_items: int | None = Field(None, alias="totalItems")



import pytest

from .attachment import transform_attachment


def test_normalize():
    attachment = {
        "type": "Document",
    }

    result = transform_attachment(attachment)

    assert result["type"] == "Document"


@pytest.mark.parametrize(
    ["media_type", "object_type"],
    [
        ("image/png", "Image"),
        ("image/jpeg", "Image"),
        ("video/mp4", "Video"),
        ("audio/mp4", "Audio"),
        ("text/plain", "Document"),
    ],
)
def test_normalize_image(media_type, object_type):
    attachment = {
        "type": "Document",
        "mediaType": media_type,
    }

    result = transform_attachment(attachment)

    assert result["type"] == object_type



"""Routines to normalize an ActivityPub activity or object.

The routines here take a dictionary and turn them into another
one."""

import logging

from muck_out.types import Activity, ActivityStub
from .object import normalize_object


logger = logging.getLogger(__name__)


def activity_stub(data: dict, actor_id: str | None = None) -> ActivityStub:
    """Builds the activity stub"""
    stub = ActivityStub.model_validate(data)

    if stub.to == [] and actor_id:
        stub.to = [actor_id]

    return stub


def normalize_activity(activity: dict, actor_id: str | None = None) -> Activity:
    """
    Normalizes activities.

    :param activity: The activity being normalized
    :param actor_id: Actor receiving this activity
    :returns:
    """
    try:
        obj = activity.get("object")
        if isinstance(obj, dict):
            try:
                obj = normalize_object(obj)
            except Exception:
                if isinstance(obj, dict):
                    obj = obj.get("id")

        stub = activity_stub(activity, actor_id=actor_id)
        dumped = stub.model_dump()
        dumped["object"] = obj

        return Activity.model_validate(dumped)

    except Exception as e:
        raise e



from muck_out.process import normalize_actor, normalize_object

actor_example = {
    "attachment": [
        {
            "type": "PropertyValue",
            "name": "Author",
            "value": "acct:helge@mymath.rocks",
        },
        {
            "type": "PropertyValue",
            "name": "Source",
            "value": "https://codeberg.org/bovine/roboherd",
        },
        {"type": "PropertyValue", "name": "Frequency", "value": "Every minute"},
    ],
    "published": "2025-09-17T17:02:51.088028",
    "@context": [
        "https://www.w3.org/ns/activitystreams",
        "https://w3id.org/security/v1",
        {
            "PropertyValue": {
                "@id": "https://schema.org/PropertyValue",
                "@context": {
                    "value": "https://schema.org/value",
                    "name": "https://schema.org/name",
                },
            }
        },
    ],
    "publicKey": {
        "id": "http://abel/actor/AFKb0cQunSBv1fC7sWbQYg#legacy-key-1",
        "owner": "http://abel/actor/AFKb0cQunSBv1fC7sWbQYg",
        "publicKeyPem": "-----BEGIN PUBLIC KEY-----\nMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEArmnl5tRTKzXoHsOmEPwk\nfC6BY1ZtueThLmXF9BjtIiHSKXvH0cema16EMb6efyHoLBvlAcdvEpP0QR3AKvb/\n48wArzIUnLi3RxOlANhUK+NHOuNJQxcPrnX06rt0MZ5K+3hK1XZQV559vecEyIk+\nxZxlz9KaSDEnsHKbHueNJvhrGV1TdppdCSe5qcml/Pkc3yIG9gq2aKx7XLIPORHe\nDRqJUpLyUX5XiOOGLHVUuKz3U87+ZWxK+GjHB46hMl+W4clJNLedxY81eGCeGoOV\ni6qEHRuWC68Q1IhexIXyujlBBHAOqgLRO/VNpg6anOZWDX/s6EvsQq4BBPNI3HZe\newIDAQAB\n-----END PUBLIC KEY-----\n",
    },
    "id": "http://abel/actor/AFKb0cQunSBv1fC7sWbQYg",
    "type": "Service",
    "inbox": "http://abel/inbox/3c2TKngdcWGJOj4j3ALC1g",
    "outbox": "http://abel/outbox/vCJcPoNbwkNlJU9o7O8qDA",
    "followers": "http://abel/followers/RKsezXFc1SGvQKvucioJxg",
    "following": "http://abel/following/CLwJywMuHXFIyPEz9sRBhg",
    "preferredUsername": "kitty",
    "name": "The kitty",
    "summary": "I'm a kitty.",
    "icon": {
        "mediaType": "image/png",
        "type": "Image",
        "url": "https://dev.bovine.social/assets/bull-horns.png",
    },
    "identifiers": ["acct:kitty@abel", "http://abel/actor/AFKb0cQunSBv1fC7sWbQYg"],
    "endpoints": {"sharedInbox": "http://abel/shared_inbox"},
    "url": "http://abel/@kitty",
}

actor = normalize_actor(actor_example)


mitra_actor = {
    "@context": [
        "https://www.w3.org/ns/activitystreams",
    ],
    "id": "http://mitra/users/admin",
    "type": "Person",
    "name": None,
    "preferredUsername": "admin",
    "inbox": "http://mitra/users/admin/inbox",
    "outbox": "http://mitra/users/admin/outbox",
    "followers": "http://mitra/users/admin/followers",
    "following": "http://mitra/users/admin/following",
    "subscribers": "http://mitra/users/admin/subscribers",
    "featured": "http://mitra/users/admin/collections/featured",
    "assertionMethod": [
        {
            "id": "http://mitra/users/admin#main-key",
            "type": "Multikey",
            "controller": "http://mitra/users/admin",
            "publicKeyMultibase": "z4MXj1wBzi9jUstyPBJ7qsa7G5waXG8McxXkjvAVHFLMxhGiJTGXK1EWseFfrUaH6Btt7TCSB1Entgg5SyAaHc1Ssh698puSozC41J2no8rgpMVRPzFVBoAYuNamM5FW9qCP1XV2y1cJ2y3gmuoreUVyn1jgW7Gb3NxsEcxjfs3fE1SzXjCnkxqVikYjkByqbXeB4AvYeXP1wjS1KB9xzKAzAjnRrui6RLHi6sCji2f7kdmxALcmjMbLyKjmYozhoG5ZqVkfNNZuAf4PLgyJQ84rDNKUGRgZeCeNh31hLLMUsSFYgpArfp813dbusq3A5BnnWN24rQtrj7nN9gQsxYjSTRhLpQcSLnwCQxJ1Sc2u9zfDwcJdW",
        },
        {
            "id": "http://mitra/users/admin#ed25519-key",
            "type": "Multikey",
            "controller": "http://mitra/users/admin",
            "publicKeyMultibase": "z6Mkgo38JzBJ6pLDyNz6YhnDN3ZXU6GcBWpp2c5dCEXcJtwW",
        },
    ],
    "publicKey": {
        "id": "http://mitra/users/admin#main-key",
        "owner": "http://mitra/users/admin",
        "publicKeyPem": "-----BEGIN PUBLIC KEY-----\nMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAqh38rZnS2RDft6RAGt5O\np4teGYSYmaeB2l5UGRCpnzcVf3m4v/x6LpXJXG8bDkNNnIZ/OzZ8tNPzKZJ6sKpI\nngayjowI/YzN8BBPJuKtaB3B4ImNrvG865OQl0h+TtfXaNTRWK8U108bMW2HH2/n\nTWJiaYiCcxLPzFRPLLd/TG6IYgeSKlRpYJ5TOdpZg1Hcx7/eRljFvdq7wmUcbM2W\nXmUYv1sjmwGQ92VCEy8d0biE/IesHdE6QM/5ALdAcQGtTn5QK7AEe43VmklkTfV+\nqDs38r3kVAMuHxnpASj+nuMp7Y7dRm92aKJ1vj93HxXTE1Ajh2eQYk8ulHQjcLky\nGwIDAQAB\n-----END PUBLIC KEY-----\n",
    },
    "generator": {
        "type": "Application",
        "implements": [
            {
                "name": "RFC-9421: HTTP Message Signatures",
                "href": "https://datatracker.ietf.org/doc/html/rfc9421",
            },
            {
                "name": "RFC-9421 signatures using the Ed25519 algorithm",
                "href": "https://datatracker.ietf.org/doc/html/rfc9421#name-eddsa-using-curve-edwards25",
            },
        ],
    },
    "manuallyApprovesFollowers": False,
    "url": "http://mitra/users/admin",
    "published": "2025-09-11T13:34:15.718558Z",
    "updated": "2025-09-11T13:34:15.718558Z",
}


sample_object = {
    "@context": [
        "https://www.w3.org/ns/activitystreams",
        {"toot": "http://joinmastodon.org/ns#", "Emoji": "toot:Emoji"},
    ],
    "id": "https://activitypub.space/post/99",
    "type": "Article",
    "to": [
        "https://www.w3.org/ns/activitystreams#Public",
        "https://activitypub.space/category/5",
    ],
    "cc": [
        "https://activitypub.space/uid/1/followers",
        "https://piefed.social/u/rimu",
        "https://community.nodebb.org/uid/2",
        "https://indieweb.social/users/phillycodehound",
        "https://mitra.social/users/silverpill",
        "https://community.nodebb.org/uid/2/followers",
        "https://indieweb.social/users/phillycodehound/followers",
        "https://mitra.social/users/silverpill/followers",
    ],
    "inReplyTo": None,
    "published": "2025-09-06T20:09:12.263Z",
    "updated": None,
    "url": "https://activitypub.space/post/99",
    "attributedTo": "https://activitypub.space/uid/1",
    "context": "https://activitypub.space/topic/18",
    "audience": "https://activitypub.space/category/5",
    "summary": " <p>Hey <a href=\"https://piefed.social/u/rimu\">@<bdi>rimu@piefed.social</bdi></a> question to you about post removal...</p> <p>If a remote user posts to a local community, and the local mod deletes it (let's say it's spam of off topic), does the local community federate a delete out?</p> <p>Technically you're not <strong>deleting</strong> the content, just removing it from the community.</p> <p>Is there a different action Piefed takes?</p>",
    "name": "Topic removal from a category/community",
    "preview": {
        "type": "Note",
        "attributedTo": "https://activitypub.space/uid/1",
        "content": "<p>Hey <a href=\"https://piefed.social/u/rimu\">@<bdi>rimu@piefed.social</bdi></a> question to you about post removal...</p>\n<p>If a remote user posts to a local community, and the local mod deletes it (let's say it's spam of off topic), does the local community federate a delete out?</p>\n<p>Technically you're not <strong>deleting</strong> the content, just removing it from the community.</p>\n<p>Is there a different action Piefed takes?</p>\n",
        "published": "2025-09-06T20:09:12.263Z",
        "attachment": [],
    },
    "content": "<p>Hey <a href=\"https://piefed.social/u/rimu\">@<bdi>rimu@piefed.social</bdi></a> question to you about post removal...</p>\n<p>If a remote user posts to a local community, and the local mod deletes it (let's say it's spam of off topic), does the local community federate a delete out?</p>\n<p>Technically you're not <strong>deleting</strong> the content, just removing it from the community.</p>\n<p>Is there a different action Piefed takes?</p>\n",
    "source": {
        "content": "Hey [rimu@piefed.social](https://activitypub.space/user/rimu%40piefed.social) question to you about post removal...\n\nIf a remote user posts to a local community, and the local mod deletes it (let's say it's spam of off topic), does the local community federate a delete out?\n\nTechnically you're not **deleting** the content, just removing it from the community.\n\nIs there a different action Piefed takes?",
        "mediaType": "text/markdown",
    },
    "tag": [
        {
            "type": "Hashtag",
            "href": "https://activitypub.space/tags/piefed",
            "name": "#piefed",
        },
        {
            "type": "Mention",
            "href": "https://piefed.social/u/rimu",
            "name": "@rimu@piefed.social",
        },
    ],
    "attachment": [],
    "replies": "https://activitypub.space/post/99/replies",
}

normalized_obj = normalize_object(sample_object)



from muck_out.types import Activity, Actor, Object, Collection

from .methods import data_to_type


def transform_activity(data: dict) -> None | Activity:
    return data_to_type(data, "activity", Activity)


def transform_actor(data: dict) -> None | Actor:
    return data_to_type(data, "actor", Actor)


def transform_object(data: dict) -> None | Object:
    return data_to_type(data, "object", Object)


def transform_embedded_object(data: dict) -> None | Object:
    return data_to_type(data, "embedded_object", Object)


def transform_collection(data: dict) -> None | Collection:
    return data_to_type(data, "collection", Collection)



from datetime import datetime

from pydantic_core import TzInfo
from muck_out.process import normalize_object
from muck_out.testing.examples import sample_object

from . import object_to_meta_info


def test_object_meta_info():
    obj = normalize_object(sample_object)

    meta = object_to_meta_info(obj)

    assert meta.id == sample_object["id"]
    assert meta.html_url == "https://activitypub.space/post/99"
    assert meta.published == datetime(2025, 9, 6, 20, 9, 12, 263000, tzinfo=TzInfo(0))



import logging
from typing import Any

from muck_out.transform.utils import sanitize_html
from muck_out.types import ObjectStub, Object

logger = logging.getLogger(__name__)


def update_for_media_type(data: dict[str, Any]) -> dict[str, Any]:
    try:
        media_type = data.get("mediaType")
        if media_type and media_type.startswith("text/markdown"):
            content = data.get("content")
            if content:
                import markdown

                copy = {**data}
                copy["content"] = markdown.markdown(content)
                return copy
    except ImportError:
        logger.warning(
            "Received markdown content; but markdown extension is not installed; Treating as html"
        )
    except Exception as e:
        logger.info(e)
    return data


def object_stub(data: dict[str, Any]) -> ObjectStub:
    """Constructs a stub from data

    This function is not a direct equivalent to ObjectStub.model_validate
    as functionality happens that is not field to field for

    - `content` filled from `contentMap`

    """

    data = update_for_media_type(data)

    stub = ObjectStub.model_validate(data)

    if stub.content is None:
        content_map = data.get("contentMap")
        if isinstance(content_map, dict):
            values = content_map.values()
            if len(values) > 0:
                stub.content = sanitize_html(list(content_map.values())[0])

    return stub


def normalize_object(obj: dict[str, Any]) -> Object:
    """Normalizes an object

    :params obj: The object to be normalized
    :returns:
    """

    stub = object_stub(obj)
    result = Object.model_validate(stub.model_dump())

    if not any([result.content, result.summary, result.name]):
        raise ValueError(
            "At least one of the properties content, summary, or name should be set"
        )

    return result



from typing import Any
from bovine.activitystreams.utils import id_for_object

from muck_out.transform import transform_to_list_of_uris
from muck_out.types import Collection, CollectionStub


def collection_stub(data: dict[str, Any]) -> CollectionStub:
    """Returns a normalized version of a collection, which possible null values"""
    result = CollectionStub.model_validate(data)

    if not result.items and data.get("orderedItems"):
        result.items = transform_to_list_of_uris(data.get("orderedItems"))

    if isinstance(data.get("first"), dict):
        first = CollectionStub.model_validate(data.get("first"))
        result.next = first.next
        result.items += first.items
        result.first = None

    return result


def convert_items(items: list[dict[str, Any] | str]) -> list[str]:
    """Converts a list of items to a list of ids

    ```pycon
    >>> convert_items([{"id": "https://site.example/1"}, "https://site.example/2"])
    ['https://site.example/1', 'https://site.example/2']

    ```
    """
    return [x for x in (id_for_object(item) for item in items) if x]


def normalize_collection(collection: dict[str, Any]) -> Collection | None:
    """Normalizes a collection"""

    stub = collection_stub(collection)

    if stub.type is None or stub.type not in (
        "Collection",
        "OrderedCollection",
        "CollectionPage",
        "OrderedCollectionPage",
    ):
        return None

    return Collection.model_validate(stub.model_dump())



from muck_out.derived.object_meta_info import ObjectMetaInfo
from muck_out.derived.util import determine_html_url
from muck_out.types import Actor, Object
from .actor_header_info import ActorHeaderInfo

__all__ = [
    "ActorHeaderInfo",
    "actor_to_header_info",
    "ObjectMetaInfo",
    "object_to_meta_info",
]


def actor_to_header_info(actor: Actor) -> ActorHeaderInfo:
    """Turns an [Actor][muck_out.types.Actor] object into a reduced version
    suitable to display in a header lien for this actor.

    ```python
    >>> from muck_out.testing.examples import actor
    >>> result = actor_to_header_info(actor)
    >>> print(result.model_dump_json(indent=2))
    {
      "id": "http://abel/actor/AFKb0cQunSBv1fC7sWbQYg",
      "avatarUrl": "https://dev.bovine.social/assets/bull-horns.png",
      "name": "The kitty",
      "identifier": "acct:kitty@abel",
      "htmlUrl": "http://abel/@kitty"
    }

    ```
    """
    avatar_url = actor.icon.get("url") if actor.icon else None
    html_url = determine_html_url(actor.url)
    return ActorHeaderInfo(
        id=actor.id,
        name=actor.name,
        identifier=actor.identifiers[0],
        avatar_url=avatar_url,
        html_url=html_url,
    )


def object_to_meta_info(obj: Object) -> ObjectMetaInfo:
    """Turns an object into its meta information


    ```python
    >>> from muck_out.testing.examples import normalized_obj
    >>> result = object_to_meta_info(normalized_obj)
    >>> print(result.model_dump_json(indent=2))
    {
      "id": "https://activitypub.space/post/99",
      "htmlUrl": "https://activitypub.space/post/99",
      "published": "2025-09-06T20:09:12.263000Z",
      "updated": null
    }

    ```
    """
    return ObjectMetaInfo(
        id=obj.id,
        html_url=determine_html_url(obj.url),
        published=obj.published,
        updated=obj.updated,
    )



import logging
from typing import Any
from urllib.parse import urlparse

from muck_out.transform.utils import remove_html
from muck_out.types import ActorStub, Actor
from muck_out.types.validated.actor import PropertyValue

logger = logging.getLogger(__name__)


def calculate_identifiers(stub: ActorStub) -> list[str]:
    if not stub.id:
        return []
    try:
        domain = urlparse(stub.id).netloc

        if stub.preferred_username:
            return [f"acct:{stub.preferred_username}@{domain}", stub.id]

        return [stub.id]
    except Exception:
        return []


def actor_stub(data: dict[str, Any]) -> ActorStub:
    """Returns the stub actor"""
    stub = ActorStub.model_validate(data)

    if data.get("identifiers") is None:
        stub.identifiers = calculate_identifiers(stub)

    return stub


def normalize_property_value(data: dict[str, Any]) -> PropertyValue | dict[str, Any]:
    if data.get("type") == "PropertyValue":
        return PropertyValue.model_validate(
            {
                "name": remove_html(data.get("name")),
                "value": remove_html(data.get("value")),
            }
        )
    return data


def normalize_actor(data: dict[str, Any]) -> Actor | None:
    """Normalizes an ActivityPub actor"""

    try:
        stub = actor_stub(data)

        if stub.inbox is None:
            return None

        if stub.identifiers is None or len(stub.identifiers) == 0:
            if stub.id is None:
                return
            stub.identifiers = [stub.id]

        return Actor.model_validate(stub.model_dump(by_alias=True))

    except Exception as e:
        logger.info(e)
        return None



from datetime import datetime
from typing import Annotated
from pydantic import BaseModel, ConfigDict, Field


class ObjectMetaInfo(BaseModel):
    """Derived information suitable to be displayed in a header line."""

    model_config = ConfigDict(
        serialize_by_alias=True,
        extra="forbid",
        field_title_generator=lambda field_name, field_info: "",
        validate_by_name=True,
    )

    id: str = Field(
        examples=["https://actor.example/alice/some/id"],
        description="id of the object, can be assumed to be globally unique.",
    )

    html_url: Annotated[
        str | None,
        Field(
            default=None,
            examples=["https://actor.example/@alice/some/ud"],
            description="Location of a html representation of the object",
            alias="htmlUrl",
        ),
    ]

    published: datetime | None = Field(
        default=None,
        description="Moment of this object being published",
    )

    updated: datetime | None = Field(
        default=None,
        description="Moment of this object being updated",
    )



from typing import Any


def list_from_value(value: Any) -> list[Any] | None:
    """Transforms a list into a value

    ```pycon
    >>> list_from_value(["aaa"])
    ['aaa']

    >>> list_from_value("aaa")
    ['aaa']

    >>> list_from_value({"a": 1})
    [{'a': 1}]

    >>> list_from_value([])

    >>> list_from_value(None)

    ```


    :returns: A list or None in case of an empty list or None as argument

    """

    if isinstance(value, list):
        if len(value) == 0:
            return None
        return value
    if isinstance(value, str) or isinstance(value, dict):
        return [value]

    return None


def transform_to_list(item) -> list:
    """Takes an item an turns it into a list"""
    list_of_items = list_from_value(item)
    if list_of_items is None:
        return []

    return list_of_items



import pytest
import json

from zipfile import ZipFile

from muck_out.types import Object, Activity

from . import normalize_data as normalize


def object_data():
    with ZipFile("test_data/samples.zip") as zip_file:
        for file_name in zip_file.namelist():
            if file_name.startswith("object/"):
                yield file_name, json.loads(zip_file.read(file_name))


def activity_data():
    with ZipFile("test_data/samples.zip") as zip_file:
        for file_name in zip_file.namelist():
            if file_name.startswith("activity/"):
                yield file_name, json.loads(zip_file.read(file_name))


failing_objects = [
    "object/object_types_12.json",
    "object/object_types_13.json",
    "object/necessary_properties_1.json",
    "object/necessary_properties_3.json",
    "object/necessary_properties_4.json",
    "object/html_bad_5.json",
    "object/html_tags_26.json",
    "object/html_tags_27.json",
    "object/html_tags_article_26.json",
    "object/html_tags_article_27.json",
    # "object/in_reply_to_4.json",
    # "object/attributed_to_2.json",
]


@pytest.mark.parametrize("file_name, data", object_data())
def test_objects(file_name, data):
    normalized = normalize(data)

    if file_name in failing_objects:
        assert normalized.object is None
    else:
        assert normalized.object

        dumped = normalized.object.model_dump()
        Object(**dumped)


@pytest.mark.parametrize("file_name, data", activity_data())
def test_activities(file_name, data):
    normalized = normalize(data)

    assert normalized.activity

    dumped = normalized.activity.model_dump()
    Activity(**dumped)


def actor_data():
    with ZipFile("test_data/funfedi-samples.zip") as zip_file:
        for file_name in zip_file.namelist():
            if file_name.startswith("actors/"):
                yield file_name, json.loads(zip_file.read(file_name))


@pytest.mark.parametrize("file_name, data", actor_data())
def test_actors(file_name, data):
    normalized = normalize(data)

    assert normalized.actor



from pydantic import BaseModel, ConfigDict, Field

from muck_out.validators import HtmlStringOrNone, IdFieldOrNone, UrlList


class CommonActor(BaseModel):
    model_config = ConfigDict(
        serialize_by_alias=True,
        validate_by_name=True,
        field_title_generator=lambda field_name, field_info: "",
    )
    inbox: IdFieldOrNone = Field(
        default=None,
        examples=["https://actor.example/inbox"],
        description="The inbox of the actor",
    )

    outbox: IdFieldOrNone = Field(
        default=None,
        examples=["https://actor.example/outbox"],
        description="The outbox of the actor",
    )

    followers: IdFieldOrNone = Field(
        default=None,
        examples=["https://actor.example/followers"],
        description="The followers collection of the actor",
    )

    following: IdFieldOrNone = Field(
        default=None,
        examples=["https://actor.example/following"],
        description="The following collection of the actor",
    )

    summary: HtmlStringOrNone = Field(
        default=None,
        examples=["My Fediverse account"],
        description="Description of the actor",
    )

    name: HtmlStringOrNone = Field(
        default=None,
        examples=["Alice"],
        description="Display name of the actor",
    )

    url: UrlList = Field(
        default=[],
        description="A list of urls that expand on the content of the object",
    )



from collections.abc import Awaitable, Callable
import json
from faststream.rabbit import RabbitBroker, RabbitExchange
from faststream import Context

from cattle_grid.dependencies import ActivityExchange
from cattle_grid.model import FetchMessage

from muck_out.cattle_grid.methods import data_to_type
from muck_out.types import Activity, Actor, Object, Collection


def fetch_method_for_key_and_type(
    broker: RabbitBroker, exchange: RabbitExchange, key: str, object_type
):
    async def fetch(actor_id: str, object_id: str):
        result = await broker.request(
            FetchMessage(actor=actor_id, uri=object_id),
            routing_key="fetch",
            exchange=exchange,
        )

        result = json.loads(result.body)

        return data_to_type(
            result,
            key,
            object_type,
        )

    return fetch


def fetch_activity_builder(
    exchange: ActivityExchange, broker: RabbitBroker = Context()
) -> Callable[[str, str], Awaitable[Actor | None]]:
    return fetch_method_for_key_and_type(broker, exchange, "activity", Activity)


def fetch_actor_builder(
    exchange: ActivityExchange, broker: RabbitBroker = Context()
) -> Callable[[str, str], Awaitable[Actor | None]]:
    return fetch_method_for_key_and_type(broker, exchange, "actor", Actor)


def fetch_collection_builder(
    exchange: ActivityExchange, broker: RabbitBroker = Context()
) -> Callable[[str, str], Awaitable[Actor | None]]:
    return fetch_method_for_key_and_type(broker, exchange, "collection", Collection)


def fetch_object_builder(
    exchange: ActivityExchange, broker: RabbitBroker = Context()
) -> Callable[[str, str], Awaitable[Actor | None]]:
    return fetch_method_for_key_and_type(broker, exchange, "object", Object)



from .activity import ActivityStub
from .actor import ActorStub
from .collection import CollectionStub
from .object import ObjectStub

__all__ = ["ObjectStub", "ActivityStub", "CollectionStub", "ActorStub"]
